from __future__ import annotations

from dataclasses import dataclass
from logging import info
from typing import Any, List

from .components import Transition
from .components.mode import FailureMode
from .components.trigger import Trigger
from .graph import StateGraph


@dataclass
class TestStep:
    """Step of a test case failing with the failure mode set.
    """
    transition: Transition
    trigger: Trigger
    mode: FailureMode

    @property
    def id(self) -> str:
        parts = [
            self.transition.from_state.id,
            self.transition.to_state.id,
            self.trigger.name,
        ]

        if self.trigger.variant:
            parts.append(self.trigger.variant)

        return '__'.join(parts + [self.mode.id])

    @property
    def fails(self) -> bool:
        return self.mode.fails

    @property
    def description(self) -> str:
        """Description of test step.
        """
        return f'{self.transition.from_state.id}: {self.mode.description}'

    @property
    def arguments(self) -> List[Any]:
        """Arguments for step calls.
        """
        return self.mode.arguments

    def __str__(self):
        return f'{self.transition.from_state.id} -> {self.transition.to_state.id} [{self.trigger.name}]; {self.mode.id}'


@dataclass
class TestCase:
    """Test case consisting out of several test steps.
    """
    steps: List[TestStep]

    @property
    def id(self):
        return self.last_step.id

    @property
    def description(self):
        return self.last_step.description

    @property
    def priority(self):
        return len(self.steps)

    @property
    def last_step(self):
        return self.steps[-1]

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other):
        return self.id == other.id

    def __str__(self):
        parts = [self.description]
        parts += [f'- {s}' for s in self.steps]
        return '\n'.join(parts)


@dataclass
class TestSuite:
    """Test suite consisting out of several test cases.
    """
    cases: List[TestCase]

    @classmethod
    def of(cls, graph: StateGraph) -> TestSuite:
        """Create new test suite out of given state graph.
        """
        cases = set()

        info('Creating test suite')

        for transitions in graph.simple_edge_paths:
            working_steps = [
                TestStep(transition, trigger=transition.default_trigger, mode=transition.default_trigger.ok)
                for transition in transitions
            ]

            for i, transition in enumerate(transitions):
                for trigger in transition.triggers or [transition.default_trigger]:
                    for mode in trigger.failure_modes.values():
                        steps = working_steps[:i] + [TestStep(transition, trigger=trigger, mode=mode)]
                        case = TestCase(steps=steps)
                        cases.add(case)

        cases = sorted(cases, key=lambda o: o.priority)

        info(f'Created {len(cases)} test cases:')
        info('')
        for case in cases:
            info(case)
            info('')

        return cls(cases)



from __future__ import annotations

from operator import attrgetter
from pathlib import Path
from typing import TYPE_CHECKING

import networkx as nx
from networkx.algorithms import shortest_path
from networkx.drawing.nx_pydot import to_pydot
from networkx.utils import pairwise
from wrapt import ObjectProxy

from .components import State, Transition

if TYPE_CHECKING:
    from typing import Optional, Tuple, Iterable, Union


class IdStrProxy(ObjectProxy):
    """Proxy that shows the id of the wrapped when stringified.
    """
    def __str__(self) -> str:
        return self.__wrapped__.id


class StateGraph:
    """Graph consisting of states and transitions.
    """
    def __init__(self) -> None:
        self.graph = nx.DiGraph()
        self.states = []
        self.transitions = []

    @property
    def initial_state(self) -> Optional[State]:
        """Get initial state of the graph.
        """
        try:
            return IdStrProxy(next(filter(attrgetter('initial'), self.states)))
        except StopIteration:
            return None

    @property
    def isolated_states(self) -> Tuple[State]:
        """Get states without any connection.
        """
        return tuple(nx.isolates(self.graph))

    @property
    def single_degree_states(self) -> Tuple[State]:
        """Get states with only a single exiting transitions.
        """
        return tuple(s for s in self.states if self.graph.out_degree(s) == 1)

    @property
    def is_closed(self) -> bool:
        """Check if a graph is closed (never ends).
        """
        return len(list(self.graph.predecessors(self.initial_state))) > 0

    @property
    def simple_edge_paths(self) -> Tuple[Tuple[Transition]]:
        """Get all paths without loops that are ending with an isolated state or with the initial node.
        """
        paths = list()

        for transition in self.transitions:
            path = [
                self.get_transition(*p)
                for p in pairwise(shortest_path(self.graph, self.initial_state, transition.from_state))
            ]
            path.append(transition)
            paths.append(path)

        return tuple(paths)

    def add_state(self, state: State) -> None:
        """Add given state to graph.
        """
        self.states.append(state)
        self.graph.add_node(IdStrProxy(state), shape='box', label=state.id)

    def add_transition(self, transition: Transition) -> None:
        """Add given transition to graph.
        """
        self.transitions.append(transition)
        self.graph.add_edge(
            IdStrProxy(transition.from_state),
            IdStrProxy(transition.to_state),
            label='|'.join([t.type for t in transition.triggers]),
        )

    def get_transition(self, from_state: State, to_state: State) -> Transition:
        """Get transition from one state to another.
        """
        for transition in self.transitions:
            if transition.from_state == from_state and transition.to_state == to_state:
                return transition

        raise ValueError('No transition found')

    def save(self, path: Path, fmt: str = 'dot') -> None:
        """Save graph to dot file.
        """
        graph = self.graph.copy()

        graph.add_node('__initial__', shape='point', width=0.2, height=0.2)
        graph.add_edge('__initial__', self.initial_state)

        dot = to_pydot(graph)

        if fmt == 'dot':
            path.write_text(dot.to_string())
        elif fmt == 'png':
            dot.write_png(path)

    @classmethod
    def of(cls, items: Iterable[Union[State, Transition]]) -> StateGraph:
        """Create new state graph ot of given states and transitions.
        """
        graph = cls()

        for item in items:
            if isinstance(item, State):
                graph.add_state(item)
            elif isinstance(item, Transition):
                try:
                    transition = graph.get_transition(item.from_state, item.to_state)
                except ValueError:
                    graph.add_transition(item)
                else:
                    transition.triggers += item.triggers

        return graph



from __future__ import annotations

from logging import INFO, basicConfig
from pathlib import Path

import click
from click import ClickException

from .generator import CodeGenerator
from .parser import ParseError, parse_file
from .suite import TestSuite

INDENT = ' ' * 4
YES = '\u2714'
NO = '\u274C'


@click.group()
def main() -> None:
    pass


@click.command('include')
@click.argument('flavor')
def include(flavor: str) -> None:
    """Get input path for given code flavor.
    """
    path = (Path(__file__).parent / 'libraries' / flavor).resolve()

    if not path.exists():
        raise ClickException(f'Unknown flavor "{flavor}"')

    click.echo((Path(__file__).parent / 'libraries' / flavor).resolve())


@click.command('graph')
@click.option('-i', '--input-file', type=click.Path(exists=False), required=True, help='Input source file')
@click.option('-o', '--output-file', type=click.Path(), required=True, help='Generated graph file')
@click.option('--fmt', type=click.Choice(['dot', 'png']), default='dot', help='File format')
def graph(input_file: str, output_file: str, fmt: str) -> None:
    """Generate dot image of graph from given input file.
    """
    input_file = Path(input_file)
    output_file = Path(output_file)

    try:
        graph = parse_file(input_file)
    except ParseError as ex:
        raise ClickException(f'{input_file}:{ex.line} {ex.message}')
    else:
        graph.save(output_file, fmt=fmt)


@click.command('analyze')
@click.option('-i', '--input-file', type=click.Path(exists=False), required=True, help='Input source file')
def analyze(input_file: str) -> None:
    """Analyze specified state machine from given input file.
    """
    def check_states():
        click.echo(f'Initial state: {graph.initial_state.id}')
        click.echo(f'States: {len(graph.states)}')

        for state in graph.states:
            click.echo(f'{INDENT}{state.id}')

    def check_transitions():
        click.echo(f'Transitions: {len(graph.transitions)}')

        for transition in graph.transitions:
            click.echo(f'{INDENT}{transition.from_state.id} -> {transition.to_state.id}')

    def check_metrics():
        click.echo(f'Closed graph: {YES if graph.is_closed else NO}')
        click.echo()

        isolated_states = graph.isolated_states
        click.echo(f'Isolated states: {YES if graph.isolated_states else NO}')

        for state in isolated_states:
            click.echo(f'{INDENT}{state.id}')

        click.echo()

        sd_states = graph.single_degree_states
        click.echo(f'Single degree states: {YES if sd_states else NO}')

        for state in sd_states:
            click.echo(f'{INDENT}{state.id}')

    input_file = Path(input_file)

    try:
        graph = parse_file(input_file)
    except ParseError as ex:
        raise ClickException(f'{input_file}:{ex.line} {ex.message}')

    check_states()
    click.echo()
    check_transitions()
    click.echo()
    check_metrics()
    click.echo()


@click.command('build')
@click.option('-i', '--input-file', type=click.Path(exists=False), required=True, help='Input source file')
@click.option('-b', '--base-file', type=click.Path(exists=False), required=True, help='Base test file')
@click.option('-o', '--output-file', type=click.Path(), required=True, help='Generated test file')
@click.option('-f', '--flavor', default='gtest', help='Flavor of generated test file')
def build(input_file: str, base_file: str, output_file: str, flavor: str) -> None:
    """Build test suite from given input file.
    """
    input_file = Path(input_file)
    base_file = Path(base_file)
    output_file = Path(output_file)

    try:
        graph = parse_file(input_file)
    except ParseError as ex:
        raise ClickException(f'{input_file}:{ex.line} {ex.message}')

    suite = TestSuite.of(graph)
    generator = CodeGenerator(flavor=flavor)
    text = generator.generate(suite, includes=[base_file.relative_to(output_file.parent)])
    output_file.write_text(text)


def run() -> None:
    """Main entry for cli.
    """
    basicConfig(
        level=INFO,
        format='%(message)s',
    )

    main.add_command(include)
    main.add_command(build)
    main.add_command(analyze)
    main.add_command(graph)
    main()



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from comment_parser.comment_parser import extract_comments_from_str
from lark import Lark, Transformer
from lark.exceptions import UnexpectedCharacters, UnexpectedEOF, UnexpectedToken

from ..components import State, Transition, Trigger
from ..graph import StateGraph
from ..utils import Resolvable

if TYPE_CHECKING:
    from typing import Tuple


class GrammarTransformer(Transformer):
    """Transformer for grammer that converts raw values to actual items.
    """
    def state(self, items) -> State:
        try:
            identifier, *options = items
        except ValueError:
            identifier = items[0]
            options = []

        return State(identifier, **dict(options))

    def transition(self, items) -> Transition:
        try:
            from_state, to_state, trigger = items
        except ValueError:
            from_state, to_state = items
            trigger = None

        if not from_state:
            return

        from_state = Resolvable.lazy(State, from_state)
        to_state = Resolvable.lazy(State, to_state)
        transition = Transition(from_state=from_state, to_state=to_state)

        if trigger:
            transition.triggers.append(trigger)

        return transition

    def trigger(self, items) -> Trigger:
        if len(items) > 1 and isinstance(items[1], str):
            typ, variant, *options = items
        else:
            typ, *options = items
            variant = None

        return Trigger.of(typ, variant=variant, **dict(options))

    def option(self, items):
        return items[0]

    def pair(self, items) -> Tuple:
        return tuple(items)

    def flag(self, items) -> Tuple:
        return (str(items[-1]), len(items) == 1)

    def implicit(self, items) -> None:
        return None

    def identifier(self, items) -> str:
        return str(items[0])

    def string(self, items) -> str:
        return str(items[0])

    def character(self, items) -> str:
        return str(items[0])

    def float(self, items) -> float:
        return float(items[0])

    def integer(self, items) -> int:
        return int(items[0])


class ParseError(RuntimeError):
    """Error indicating a parsing problem.
    """
    def __init__(self, message: str, line: int):
        self.message = message
        self.line = line


def parse(text: str) -> StateGraph:
    """Parse given source code text into a state graph.
    """
    grammer_path = Path(__file__).parent / 'grammar.lark'
    parser = Lark(grammer_path.read_text())
    transformer = GrammarTransformer()

    definitions = []

    for comment in extract_comments_from_str(text, mime='text/x-c'):
        comment_lines = [line.strip(' *') for line in comment.text().split('\n')]

        for lineno, line in enumerate(comment_lines):
            lineno += comment.line_number()

            if not line.startswith('@'):
                continue

            try:
                ast = parser.parse(line)
                element = transformer.transform(ast)
                definitions.append(element)
            except (UnexpectedCharacters, UnexpectedToken) as ex:
                if ex.column > 1:
                    message = ex.args[0].split('at')[0]
                    raise ParseError(message=message, line=lineno)
            except UnexpectedEOF:
                raise ParseError(message='Unexpected end', line=lineno)

    Resolvable.resolve_all(definitions)
    graph = StateGraph.of(definitions)
    return graph


def parse_file(file_path: Path) -> StateGraph:
    """Parse file at given path into a state graph.
    """
    return parse(file_path.read_text())



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import TYPE_CHECKING

from .mode import FailureModeResolver, failure_mode

if TYPE_CHECKING:
    from .mode import FailureMode

    from typing import List, Optional, Dict

TRIGGER_IMPLS = {}


@dataclass
class Trigger(FailureModeResolver, ABC):
    """Base trigger class. All implementations must inherit this class and specify a short name.
    """
    variant: Optional[str] = None

    def __init_subclass__(cls, *, short_name: str):
        TRIGGER_IMPLS[short_name] = cls
        cls.name = short_name

    @property
    def type(self) -> str:
        typ = self.name or ''

        if typ and self.variant:
            typ += '#' + self.variant

        return typ

    @property
    def failure_modes(self) -> Dict[str, FailureMode]:
        """Get dict of all failure modes specified in impl class.
        """
        return {k: m for k, m in self._failure_modes.items() if m.possible}

    @staticmethod
    def of(name: str, **kwargs):
        """Create instance of registered impl class.
        """
        trigger_cls = TRIGGER_IMPLS[name]
        return trigger_cls(**kwargs)


@dataclass
class DefaultTrigger(Trigger, short_name='default'):
    @failure_mode(fails=False)
    def ok(self) -> List[None]:
        """Transition is triggered.
        """
        return [None]

    @failure_mode
    def no(self) -> List[None]:
        """Transition is not triggered.
        """
        return []


@dataclass
class TimeoutTrigger(Trigger, short_name='timeout'):
    """Trigger type which is coupled with a specified timeout.
    """
    min_factor: int = 0
    max_factor: int = 2
    operator: str = '>'

    @failure_mode(fails=False)
    def ok(self) -> List[int]:
        """Timeout matches expected value.
        """
        if not self.earlier.fails:
            return self.earlier.arguments
        else:
            return self.later.arguments

    @failure_mode(fails=lambda self: '=' not in self.operator)
    def exact(self) -> List[int]:
        """Timeout at exact threshold.
        """
        return [1]

    @failure_mode(fails=lambda self: '>' in self.operator)
    def earlier(self) -> List[int]:
        """Timeout less than expected.
        """
        return [self.min_factor]

    @failure_mode(fails=lambda self: '<' in self.operator)
    def later(self) -> List[int]:
        """Timeout greater than expected.
        """
        return [self.max_factor]


@dataclass
class ReceiveTrigger(Trigger, short_name='receive'):
    """Trigger waiting for an incoming message.
    """
    max_count: int = 3

    @failure_mode(fails=False)
    def ok(self) -> List[int]:
        """Message is received as expected.
        """
        return [1]

    @failure_mode
    def no(self) -> List[int]:
        """No message is received.
        """
        return []

    @failure_mode(requires=lambda self: self.max_count > 1)
    def more(self) -> List[int]:
        """More messages are received than expected.
        """
        return list(range(1, self.max_count + 1))



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from caseconverter import pascalcase
from jinja2 import FileSystemLoader
from jinja2.environment import Environment

from .suite import TestSuite

if TYPE_CHECKING:
    from typing import Iterable, Optional


class CodeGenerator:
    """Generate test source code out of given test suite.
    """
    def __init__(self, flavor: str = 'gtest') -> None:
        """
        Args:
            flavor (str, optional): Code flavor. Defaults to 'gtest'.
        """
        path = Path(__file__).parent / 'templates' / flavor

        if not path.exists():
            raise ValueError(f'Unknown flavor {flavor}')

        self.env = Environment(trim_blocks=True, lstrip_blocks=True)
        self.env.loader = FileSystemLoader(path)
        self.env.filters['pascalcase'] = pascalcase

    def generate(self, element: TestSuite, includes: Optional[Iterable] = None) -> str:
        """Generate code out of given test suite and add given includes.

        Args:
            element (TestSuite): Test suite containing test cases for generation.
            includes (Optional[Iterable], optional): List of additional files to include. Defaults to None.

        Returns:
            str: Generated source code
        """
        template = self.env.get_template('suite.j2')
        includes = set(includes) if includes else set()
        return template.render(
            includes=includes,
            suite=element,
        )
